"""
CarlaPy
S[pressed] + arrow keys: adjust view position relative to target
X[pressed] + arrow keys: adjust view pitch and yaw
T: change attachment type
"""

import math
import weakref
import carla
import threading
import numpy as np
import os
os.environ['PYGAME_HIDE_SUPPORT_PROMPT'] = "1"
import pygame
from . import carla_utils


class HelpText:
    def __init__(self, width, height):
        lines = __doc__.strip('\n').split('\n')
        font_size = 24
        line_space = 30
        padding_up = 30
        padding_down = 20
        padding_left = 20
        padding_right = 20
        try:
            font = pygame.font.SysFont("notomono", font_size)
        except RuntimeError:
            print('using default font')
            font = pygame.font.Font(None, font_size)
        text_width = max(font.render(line, True, (255, 255, 255)).get_width() for line in lines)
        text_height = line_space * (len(lines) - 1) + font_size
        text_surface = pygame.Surface((text_width, text_height))
        for i, line in enumerate(lines):
            text_surface.blit(font.render(line, True, (255, 255, 255)), (0, i * line_space))
        surface_dim = (padding_left + padding_right + text_width, padding_up + padding_down + text_height)
        self.surface = pygame.Surface(surface_dim)
        self.surface.fill((0, 0, 0, 0))
        self.surface.blit(text_surface, (padding_left, padding_up))
        self.surface.set_alpha(200)
        self.pos = (max((width - surface_dim[0]) // 2, 0), max((height - surface_dim[1]) // 2, 0))
        self.show = False

    def toggle(self):
        self.show = not self.show

    def render(self, target_surface):
        if self.show:
            target_surface.blit(self.surface, self.pos)

class SensorManager:
    def __init__(self,
                 parent: carla.Actor,
                 client: carla.Client,
                 view_width,
                 view_height,
                 sensor_update_event,
                 ):
        self.stop_event = threading.Event()
        self.parent = parent
        self.client = client
        self.view_width = view_width
        self.view_height = view_height
        self.sensor_update_event = sensor_update_event
        self.sensor_index = 0
        self.camera_rgb = {
            'pattern': 'sensor.camera.rgb',
            'desc': 'Camera RGB',
            'attributes': {
                'role_name': 'spy_spectator',
                'sensor_tick': 0.04,
                'image_size_x': self.view_width,
                'image_size_y': self.view_height,
                'fov': 90,
            },
            'color_converter': carla.ColorConverter.Raw,
        }
        self.sensors = [
            self.camera_rgb,
        ]
        self.sensor: carla.Sensor | None = None
        self.bp: carla.ActorBlueprint | None = None
        self.transform: carla.Transform = carla.Transform(carla.Location(), carla.Rotation())
        self.attachment: carla.AttachmentType = carla.AttachmentType.Rigid

        self.surface: pygame.Surface | None = None

    def stop(self):
        print('SensorManager: stop')
        self.stop_event.set()
        if self.sensor is not None and self.sensor.is_alive:
            if self.sensor.is_listening:
                self.sensor.stop()
            self.sensor.destroy()
            self.sensor = None

    def move(self, keys):
        if self.attachment != carla.AttachmentType.Rigid:
            print('current attachment type is {}'.format(self.attachment))
            print('move is only available for {}'.format(carla.AttachmentType.Rigid))
            return
        move = 0.1
        degree = 1
        if keys[pygame.K_LSHIFT]:
            move = 0.01
            degree = 0.1
        if keys[pygame.K_s]:
            if keys[pygame.K_UP]:
                self.transform.location.x += move
            if keys[pygame.K_DOWN]:
                self.transform.location.x -= move
            if keys[pygame.K_LEFT]:
                self.transform.location.y -= move
            if keys[pygame.K_RIGHT]:
                self.transform.location.y += move
            if keys[pygame.K_a]:
                self.transform.location.z += move
            if keys[pygame.K_d]:
                self.transform.location.z -= move
        if keys[pygame.K_x]:
            if keys[pygame.K_UP]:
                self.transform.rotation.pitch += degree
            if keys[pygame.K_DOWN]:
                self.transform.rotation.pitch -= degree
            if keys[pygame.K_LEFT]:
                self.transform.rotation.yaw -= degree
            if keys[pygame.K_RIGHT]:
                self.transform.rotation.yaw += degree
        self.sensor.set_transform(self.transform)
        print('Sensor move to:', self.transform)

    def spawn(self):
        if self.sensor is not None and self.sensor.is_alive:
            if self.sensor.is_listening:
                self.sensor.stop()
            self.sensor.destroy()
            self.sensor = None
        if self.sensor_index >= len(self.sensors):
            self.sensor_index = 0
        sensor_data = self.sensors[self.sensor_index]
        self.bp = carla_utils.get_blueprint(sensor_data['pattern'], self.client)
        attributes = sensor_data.get('attributes', {})
        for attr, value in attributes.items():
            self.bp.set_attribute(attr, str(value))
        if math.isfinite(self.parent.bounding_box.location.x):
            self.transform.location.x = -self.parent.bounding_box.extent.x * 1.88
            self.transform.location.z = self.parent.bounding_box.extent.z * 3.76
            self.transform.rotation.pitch = -20
        self.sensor = self.client.get_world().spawn_actor(self.bp, self.transform, self.parent, self.attachment)
        weak_self = weakref.ref(self)
        self.sensor.listen(lambda image: SensorManager.callback(weak_self, image))

    @staticmethod
    def callback(weak_self, image):
        self = weak_self()
        if not self:
            return
        if self.stop_event.is_set():
            return
        sensor_data = self.sensors[self.sensor_index]
        if sensor_data.get('color_converter', None) is not None:
            image.convert(sensor_data['color_converter'])
        if self.sensor_index == 0:
            array = np.frombuffer(image.raw_data, dtype=np.uint8)
            array = np.reshape(array, (image.height, image.width, 4))
            array = array[:, :, :3]
            array = array[:, :, ::-1]
            self.surface = pygame.surfarray.make_surface(array.swapaxes(0, 1))
            pygame.event.post(pygame.event.Event(self.sensor_update_event))

class CarlaClient:
    def __init__(self, actor_id, client: carla.Client):
        self.client = client
        self.client.get_world().wait_for_tick()
        self.actor = carla_utils.get_actor(actor_id, self.client)
        if self.actor is None:
            raise RuntimeError('Actor {} not found'.format(actor_id))
        if not self.actor.is_alive:
            raise RuntimeError('Actor {} not alive'.format(actor_id))

    def gui_title(self):
        return f"Actor(id={self.actor.id}, role_name={self.actor.attributes.get('role_name', 'None')}, type={self.actor.type_id})"

class GUI:
    def __init__(self, actor_id, client: carla.Client):
        self.stop_event = threading.Event()
        self.logic_width = 800
        self.logic_height = 600
        pygame.init()
        pygame.key.set_repeat(300, 50)
        self.logic_surface = pygame.Surface((self.logic_width, self.logic_height))
        self.w = self.logic_width
        self.h = self.logic_height
        self.ratio = 1.
        self.display = None
        self.init_display()

        self.carla_client = CarlaClient(actor_id, client)
        pygame.display.set_caption(self.carla_client.gui_title())

        self.sensor_update_event = pygame.USEREVENT + 1
        self.sensor_manager = SensorManager(self.carla_client.actor,
                                            client,
                                            self.logic_width,
                                            self.logic_height,
                                            self.sensor_update_event,
                                            )
        self.sensor_manager.spawn()

        self.help_text = HelpText(self.logic_width, self.logic_height)

    def stop(self):
        print('GUI: stop')
        self.stop_event.set()
        self.sensor_manager.stop()

    def init_display(self):
        self.display = pygame.display.set_mode((self.w, self.h),
                                               pygame.DOUBLEBUF |
                                               pygame.HWSURFACE,
                                               )

    # def flip_display(self, surface, destination = (0, 0)):
    #     self.display.blit(surface, destination)
    #     pygame.display.flip()

    # def update_display(self):
    #     self.display.blit(pygame.transform.scale(self.logic_surface, (self.w, self.h)), (0, 0))
    #     pygame.display.flip()

    def resize_display(self, width, height):
        if width < self.w or height < self.h:
            self.ratio = min(width / self.w, height / self.h)
            self.w = int(self.w * self.ratio)
            self.h = int(self.h * self.ratio)
            self.init_display()
            # self.update_display()

    def run(self):
        try:
            while not self.stop_event.is_set():
                event = pygame.event.wait()
                match event.type:
                    case pygame.QUIT:
                        self.stop()
                        break
                    case pygame.VIDEORESIZE:
                        # self.resize_display(event.w, event.h)
                        pass
                    case self.sensor_update_event:
                        if self.sensor_manager.surface is not None:
                            self.display.blit(self.sensor_manager.surface, (0, 0))
                    case pygame.KEYDOWN:
                        keys = pygame.key.get_pressed()
                        if keys[pygame.K_s] or keys[pygame.K_x]:
                            self.sensor_manager.move(keys)
                        if keys[pygame.K_h]:
                            self.help_text.toggle()
                self.help_text.render(self.display)
                pygame.display.flip()
        except RuntimeError as e:
            print(e)
            self.stop()
        except KeyboardInterrupt:
            self.stop()



import math
import time
import carla


def get_actors(client: carla.Client) -> carla.ActorList:
    return client.get_world().get_actors()

def get_actor(actor_id, client: carla.Client) -> carla.Actor | None:
    if actor_id is None:
        return None
    try:
        actor_id = int(actor_id)
    except ValueError:
        pass
    try:
        if isinstance(actor_id, int):
            return client.get_world().get_actor(actor_id)
        if len(actor_id) == 0:
            return None
        actors = get_actors(client)
        for actor in actors:
            if actor.attributes.get('role_name', '') == actor_id:
                return actor
            if actor.type_id == actor_id:
                return actor
    except RuntimeError:
        pass
    return None

def is_vehicle(actor: carla.Actor) -> bool:
    return 'vehicle' in actor.type_id

def is_sensor(actor: carla.Actor) -> bool:
    return 'sensor' in actor.type_id

def print_simple_actor(actor: carla.Actor, prefix: str = '', max_len = 0):
    print(prefix, f"Actor(id={actor.id}, role_name={actor.attributes.get('role_name', 'None'):<{max_len}}, type={actor.type_id})", sep='')

def print_actor_basic(
        actor: carla.Actor,
        client: carla.Client,
        prefix: str = '',
        print_parent: bool = True,
        print_sons: bool = True,
        print_attributes: bool = True,
        print_location: bool = True,
        print_rotation: bool = True,
):
    actors = get_actors(client)
    max_len = max([len(a.attributes.get('role_name', '')) for a in actors])
    print_simple_actor(actor, prefix=prefix, max_len=max_len)
    if print_parent and actor.parent is not None:
        print_actor_basic(actor.parent,
                          client,
                          prefix + 'Parent: ',
                          print_parent = False,
                          print_sons = is_vehicle(actor.parent),
                          print_attributes = False,
                          print_location = False,
                          print_rotation = False,
                          )
    if print_sons:
        for a in actors:
            if a.parent is None or a.parent.id != actor.id:
                continue
            print_actor_basic(a,
                              client,
                              prefix + 'Son: ',
                              print_parent = False,
                              print_sons = False,
                              print_attributes = False,
                              print_location = False,
                              print_rotation = False,
                              )
    if print_attributes:
        print(prefix + 'Attributes:')
        max_len = max(len(attr) for attr in actor.attributes)
        for attr in actor.attributes:
            print(' ' * (len(prefix) + 4), f"{attr:<{max_len}}", ': ', actor.attributes[attr], sep='')
    if print_location:
        print(prefix, 'Transform: ', actor.get_location(), sep='')
    if print_rotation:
        print(prefix, 'Transform: ', actor.get_transform().rotation, sep='')

def print_actor_move(actor: carla.Actor, prefix: str = ''):
    print(prefix, 'Velocity       : ', actor.get_velocity(), sep='')
    print(prefix, 'Acceleration   : ', actor.get_acceleration(), sep='')
    print(prefix, 'AngularVelocity: ', actor.get_angular_velocity(), sep='')

def print_actor_boundbox(actor: carla.Actor, prefix: str = ''):
    bb = actor.bounding_box
    print(prefix, 'BoundingBox Location: ', bb.location, sep='')
    print(prefix, 'BoundingBox Extent  : ', bb.extent, sep='')
    print(prefix, 'BoundingBox Rotation: ', bb.rotation, sep='')

def print_actor_vehicle_wheel(actor: carla.Actor, prefix: str = ''):
    if not is_vehicle(actor):
        return
    vehicle: carla.Vehicle = actor
    print(prefix, 'FL_Wheel: ', vehicle.get_wheel_steer_angle(carla.VehicleWheelLocation.FL_Wheel), sep='')
    print(prefix, 'FR_Wheel: ', vehicle.get_wheel_steer_angle(carla.VehicleWheelLocation.FR_Wheel), sep='')

def print_actor_vehicle_control(actor: carla.Actor, prefix: str = ''):
    if not is_vehicle(actor):
        return
    vehicle: carla.Vehicle = actor
    control = vehicle.get_control()
    print(prefix, 'Control Throttle  : ', control.throttle, sep='')
    print(prefix, 'Control Brake     : ', control.brake, sep='')
    print(prefix, 'Control Steer     : ', control.steer, sep='')
    print(prefix, 'Control Reverse   : ', control.reverse, sep='')
    print(prefix, 'Control HandBrake : ', control.hand_brake, sep='')
    print(prefix, 'Control Gear      : ', control.gear, sep='')
    print(prefix, 'Control ManualGear: ', control.manual_gear_shift, sep='')

def print_actor_vehicle_physics(actor: carla.Actor, prefix: str = ''):
    if not is_vehicle(actor):
        return
    vehicle: carla.Vehicle = actor
    physics = vehicle.get_physics_control()
    print(prefix, 'Vehicle Physics:', sep='')
    p = ' ' * (len(prefix) + 4)
    pp = ' ' * (len(prefix) + 8)
    print(p, 'moi             : ', physics.moi, sep='')
    print(p, 'mass            : ', physics.mass, sep='')
    print(p, 'max rpm         : ', physics.max_rpm, sep='')
    print(p, 'final_ratio     : ', physics.final_ratio, sep='')
    print(p, 'center_of_mass  : ', physics.center_of_mass, sep='')
    print(p, 'clutch_strength : ', physics.clutch_strength, sep='')
    print(p, 'use_gear_autobox: ', physics.use_gear_autobox, sep='')
    print(p, 'gear_switch_time: ', physics.gear_switch_time, sep='')
    print(p, 'drag_coefficient: ', physics.drag_coefficient, sep='')
    print(p, 'use_sweep_wheel_collision                   : ', physics.use_sweep_wheel_collision, sep='')
    print(p, 'damping_rate_full_throttle                  : ', physics.damping_rate_full_throttle, sep='')
    print(p, 'damping_rate_zero_throttle_clutch_engaged   : ', physics.damping_rate_zero_throttle_clutch_engaged, sep='')
    print(p, 'damping_rate_zero_throttle_clutch_disengaged: ', physics.damping_rate_zero_throttle_clutch_disengaged, sep='')
    print(p, 'Torque Curve:', sep='')
    for curve in physics.torque_curve:
        print(pp, curve, sep='')
    print(p, 'Forward Gears:', sep='')
    for gear in physics.forward_gears:
        print(pp, gear, sep='')
    print(p, 'Steering Curve:', sep='')
    for curve in physics.steering_curve:
        print(pp, curve, sep='')
    print(p, 'Wheel:', sep='')
    for i, wheel in enumerate(physics.wheels):
        print(pp, i, ': tire_friction       : ', wheel.tire_friction, sep='')
        print(pp, i, ': damping_rate        : ', wheel.damping_rate, sep='')
        print(pp, i, ': max_steer_angle     : ', wheel.max_steer_angle, sep='')
        print(pp, i, ': radius              : ', wheel.radius, sep='')
        print(pp, i, ': max_brake_torque    : ', wheel.max_brake_torque, sep='')
        print(pp, i, ': max_handbrake_torque: ', wheel.max_handbrake_torque, sep='')
        print(pp, i, ': position            : ', wheel.position, sep='')
        print(pp, i, ': long_stiff_value    : ', wheel.long_stiff_value, sep='')
        print(pp, i, ': lat_stiff_max_load  : ', wheel.lat_stiff_max_load, sep='')
        print(pp, i, ': lat_stiff_value     : ', wheel.lat_stiff_value, sep='')
        print()

def print_actor(
        actor: carla.Actor,
        client: carla.Client,
        prefix: str = '',
        print_parent: bool = True,
        print_sons: bool = True,
        print_attributes: bool = True,
        print_location: bool = True,
        print_rotation: bool = True,
        print_move: bool = True,
        print_boundbox: bool = True,
        print_vehicle_wheel: bool = True,
        print_vehicle_control: bool = True,
        print_vehicle_physics: bool = True,
):
    print_actor_basic(actor, client, prefix, print_parent, print_sons, print_attributes, print_location, print_rotation)
    if print_move:
        print()
        print_actor_move(actor, prefix)
    if print_boundbox and math.isfinite(actor.bounding_box.location.x):
        print()
        print_actor_boundbox(actor, prefix)
    if print_vehicle_wheel and is_vehicle(actor):
        print()
        print_actor_vehicle_wheel(actor, prefix)
    if print_vehicle_control and is_vehicle(actor):
        print()
        print_actor_vehicle_control(actor, prefix)
    if print_vehicle_physics and is_vehicle(actor):
        print()
        print_actor_vehicle_physics(actor, prefix)

def print_actors_tree(actors: carla.ActorList, pattern: str = ''):
    root_ids: set[int] = set()
    pattern = pattern.strip('*')
    if pattern:
        for actor in actors:
            if pattern in actor.attributes.get('role_name', '') or pattern in actor.type_id:
                while actor.parent is not None:
                    actor = actors.find(actor.parent.id)
                root_ids.add(actor.id)
    if not pattern.startswith('*'):
        pattern = '*' + pattern
    if not pattern.endswith('*'):
        pattern = pattern + '*'
    for actor in actors.filter(pattern):
        while actor.parent is not None:
            actor = actors.find(actor.parent.id)
        root_ids.add(actor.id)
    max_len = max([len(a.attributes.get('role_name', '')) for a in actors])
    def print_tree(a: carla.Actor, prefix: str = ''):
        print_simple_actor(a, prefix, max_len)
        for actor in actors:
            if actor.parent is not None and actor.parent.id == a.id:
                print_tree(actor, prefix + ' ' * 4)
    for actor_id in root_ids:
        print_tree(actors.find(actor_id))

def destroy_actor(actor_id, client: carla.Client):
    try:
        while True:
            client.get_world().wait_for_tick()
            actor = get_actor(actor_id, client)
            if actor is None or not actor.is_alive:
                return
            actors = client.get_world().get_actors()
            for a in actors:
                if a.parent is not None and a.parent.id == actor.id and a.is_alive:
                    print('destroying son actor: ')
                    print_simple_actor(a)
                    a.destroy()
            print('destroying actor: ')
            print_simple_actor(actor)
            actor.destroy()
    except RuntimeError as e:
        print(e)
        print('check the actor {} status'.format(actor_id))
        pass

def print_settings(settings: carla.WorldSettings):
    print('synchronous_mode      : ', settings.synchronous_mode)
    print('no_rendering_mode     : ', settings.no_rendering_mode)
    print('fixed_delta_seconds   : ', settings.fixed_delta_seconds)
    print('substepping           : ', settings.substepping)
    print('max_substep_delta_time: ', settings.max_substep_delta_time)
    print('max_substeps          : ', settings.max_substeps)
    print('max_culling_distance  : ', settings.max_culling_distance)
    print('deterministic_ragdolls: ', settings.deterministic_ragdolls)
    print('tile_stream_distance  : ', settings.tile_stream_distance)
    print('actor_active_distance : ', settings.actor_active_distance)
    print('spectator_as_ego      : ', settings.spectator_as_ego)

def get_blueprint(pattern, client: carla.Client) -> carla.ActorBlueprint | None:
    if len(pattern) == 0:
        return None
    try:
        return client.get_world().get_blueprint_library().find(pattern)
    except RuntimeError:
        pass
    if not pattern.startswith('*'):
        pattern = '*' + pattern
    if not pattern.endswith('*'):
        pattern = pattern + '*'
    bps = client.get_world().get_blueprint_library().filter(pattern)
    if len(bps) == 0:
        return None
    return bps[0]

def spawn_actor(
        client: carla.Client,
        bp: carla.ActorBlueprint,
        transform: carla.Transform = None,
        parent: carla.Actor = None,
        attachment_type = carla.AttachmentType.Rigid
):
    if transform is None:
        loc = carla.Location()
        rot = carla.Rotation()
        transform = carla.Transform(loc, rot)
    if parent is None:
        return client.get_world().spawn_actor(bp, transform)
    return client.get_world().spawn_actor(bp, transform, parent, attachment_type)

def get_transform(x, y, z, roll, pitch, yaw) -> carla.Transform | None:
    if {x, y, z, roll, pitch, yaw} == {None}:
        return None
    x = 0 if x is None else x
    y = 0 if y is None else y
    z = 0 if z is None else z
    roll = 0 if roll is None else roll
    pitch = 0 if pitch is None else pitch
    yaw = 0 if yaw is None else yaw
    return carla.Transform(carla.Location(x, y, z), carla.Rotation(pitch, yaw, roll))

def move_to(
        client: carla.Client, actor_id: str = None, target_id: str = None,
        x: float = None, y: float = None, z: float = None,
        roll: float = None, pitch: float = None, yaw: float = None,
):
    client.get_world().wait_for_tick()
    actor = get_actor(actor_id, client)
    if actor is None:
        return
    target = get_actor(target_id, client)
    t = get_transform(x, y, z, roll, pitch, yaw)
    if target is None:
        if t is None:
            return
        actor.set_transform(t)
        return
    if t is None:
        actor.set_transform(target.get_transform())
        return
    dummy_bp = get_blueprint('sensor.other.collision', client)
    dummy_bp.set_attribute('role_name', 'dummy_for_move_to')
    dummy = spawn_actor(client, dummy_bp, t, target, carla.AttachmentType.Rigid)
    client.get_world().wait_for_tick()
    actor.set_transform(dummy.get_transform())
    dummy.destroy()



import functools
import sys
import os
import fcntl
from typing import Callable

def safe_invoke() -> Callable:
    def decorator(fn: Callable) -> Callable:
        @functools.wraps(fn)
        def wrapper(*args, **kwargs):
            try:
                return fn(*args, **kwargs)
            except RuntimeError as e:
                print(e)
                return None
        return wrapper
    return decorator

def set_stdout_nonblocking():
    fd = sys.stdout.fileno()
    try:
        flags = fcntl.fcntl(fd, fcntl.F_GETFL)
        fcntl.fcntl(fd, fcntl.F_SETFL, flags | os.O_NONBLOCK)
    finally:
        pass



from __future__ import annotations

from typing import Any
import numpy as np

import carla

class Response:
	"""
	States the result of executing a command as either the ID of the actor to whom the command was applied to (when succeeded) or an error string (when failed).  actor ID, depending on whether or not the command succeeded. The method apply_batch_sync() in carla.Client returns a list of these to summarize the execution of a batch.
	"""

	actor_id: int
	"""Actor to whom the command was applied to. States that the command was successful."""

	error: str
	"""A string stating the command has failed."""

	def has_error(self) -> bool:
		"""
		Returns True if the command execution fails, and False if it was successful.

		:return: bool
		"""
		...



class SpawnActor:
	"""
	Command adaptation of spawn_actor() in carla.World. Spawns an actor into the world based on the blueprint provided and the transform. If a parent is provided, the actor is attached to it.
	"""

	transform: carla.Transform
	"""Transform to be applied."""

	parent_id: int
	"""Identificator of the parent actor."""

	def __init__(self):
		"""

		"""
		...

	def __init__(self, blueprint: carla.ActorBlueprint, transform: carla.Transform):
		"""


		:param blueprint: (carla.ActorBlueprint) 

		:param transform: (carla.Transform) 
		"""
		...

	def __init__(self, blueprint: carla.ActorBlueprint, transform: carla.Transform, parent: str):
		"""


		:param blueprint: (carla.ActorBlueprint) 

		:param transform: (carla.Transform) 

		:param parent: (str) 
		"""
		...

	def then(self, command: str):
		"""
		Links another command to be executed right after. It allows to ease very common flows such as spawning a set of vehicles by command and then using this method to set them to autopilot automatically.

		:param command: (str) a Carla command.
		"""
		...



class DestroyActor:
	"""
	Command adaptation of destroy() in carla.Actor that tells the simulator to destroy this actor. It has no effect if the actor was already destroyed. When executed with apply_batch_sync() in carla.Client there will be a command.Response that will return a boolean stating whether the actor was successfully destroyed.
	"""

	actor_id: int
	"""Actor affected by the command"""

	def __init__(self, actor: str):
		"""


		:param actor: (str) Actor or its ID to whom the command will be applied to.
		"""
		...



class ApplyVehiclePhysicsControl:
	"""
	Command adaptation of apply_physics_control() in carla.Vehicle. Applies a new physics control to a vehicle, modifying its physical parameters.
	"""

	actor_id: int
	"""Vehicle actor affected by the command."""

	physics_control: carla.VehiclePhysicsControl
	"""Physics control to be applied."""

	def __init__(self, actor: str, physics_control: carla.VehiclePhysicsControl):
		"""


		:param actor: (str) Actor or its ID to whom the command will be applied to.

		:param physics_control: (carla.VehiclePhysicsControl) 
		"""
		...



class ApplyVehicleControl:
	"""
	Command adaptation of apply_control() in carla.Vehicle. Applies a certain control to a vehicle.
	"""

	actor_id: int
	"""Vehicle actor affected by the command."""

	control: carla.VehicleControl
	"""Vehicle control to be applied."""

	def __init__(self, actor: str, control: carla.VehicleControl):
		"""


		:param actor: (str) Actor or its ID to whom the command will be applied to.

		:param control: (carla.VehicleControl) 
		"""
		...



class ApplyVehicleAckermannControl:
	"""
	Command adaptation of apply_ackermann_control() in carla.Vehicle. Applies a certain akermann control to a vehicle.
	"""

	actor_id: int
	"""Vehicle actor affected by the command."""

	control: carla.AckermannVehicleControl
	"""Vehicle ackermann control to be applied."""

	def __init__(self, actor: str, control: carla.AckermannVehicleControl):
		"""


		:param actor: (str) Actor or its ID to whom the command will be applied to.

		:param control: (carla.AckermannVehicleControl) 
		"""
		...



class ApplyWalkerControl:
	"""
	Command adaptation of apply_control() in carla.Walker. Applies a control to a walker.
	"""

	actor_id: int
	"""Walker actor affected by the command."""

	control: carla.WalkerControl
	"""Walker control to be applied."""

	def __init__(self, actor: str, control: carla.WalkerControl):
		"""


		:param actor: (str) Actor or its ID to whom the command will be applied to.

		:param control: (carla.WalkerControl) 
		"""
		...



class ApplyTransform:
	"""
	Command adaptation of set_transform() in carla.Actor. Sets a new transform to an actor.
	"""

	actor_id: int
	"""Actor affected by the command."""

	transform: carla.Transform
	"""Transformation to be applied."""

	def __init__(self, actor: str, transform: carla.Transform):
		"""


		:param actor: (str) Actor or its ID to whom the command will be applied to.

		:param transform: (carla.Transform) 
		"""
		...



class ApplyWalkerState:
	"""
	Apply a state to the walker actor. Specially useful to initialize an actor them with a specific location, orientation and speed.
	"""

	actor_id: int
	"""Walker actor affected by the command."""

	transform: carla.Transform
	"""Transform to be applied."""

	speed: float
	"""Speed to be applied."""

	def __init__(self, actor: str, transform: carla.Transform, speed: float):
		"""


		:param actor: (str) Actor or its ID to whom the command will be applied to.

		:param transform: (carla.Transform) 

		:param speed: (float) 
		"""
		...



class ApplyTargetVelocity:
	"""
	Command adaptation of set_target_velocity() in carla.Actor.
	"""

	actor_id: int
	"""Actor affected by the command."""

	velocity: carla.Vector3D
	"""The 3D velocity applied to the actor."""

	def __init__(self, actor: str, velocity: carla.Vector3D):
		"""


		:param actor: (str) Actor or its ID to whom the command will be applied to.

		:param velocity: (carla.Vector3D) Velocity vector applied to the actor.
		"""
		...



class ApplyTargetAngularVelocity:
	"""
	Command adaptation of set_target_angular_velocity() in carla.Actor. Sets the actor's angular velocity vector.
	"""

	actor_id: int
	"""Actor affected by the command."""

	angular_velocity: carla.Vector3D
	"""The 3D angular velocity that will be applied to the actor."""

	def __init__(self, actor: str, angular_velocity: carla.Vector3D):
		"""


		:param actor: (str) Actor or its ID to whom the command will be applied to.

		:param angular_velocity: (carla.Vector3D) Angular velocity vector applied to the actor.
		"""
		...



class ApplyImpulse:
	"""
	Command adaptation of add_impulse() in carla.Actor. Applies an impulse to an actor.
	"""

	actor_id: int
	"""Actor affected by the command."""

	impulse: carla.Vector3D
	"""Impulse applied to the actor."""

	def __init__(self, actor: str, impulse: carla.Vector3D):
		"""


		:param actor: (str) Actor or its ID to whom the command will be applied to.

		:param impulse: (carla.Vector3D) 
		"""
		...



class ApplyForce:
	"""
	Command adaptation of add_force() in carla.Actor. Applies a force to an actor.
	"""

	actor_id: int
	"""Actor affected by the command."""

	force: carla.Vector3D
	"""Force applied to the actor over time."""

	def __init__(self, actor: str, force: carla.Vector3D):
		"""


		:param actor: (str) Actor or its ID to whom the command will be applied to.

		:param force: (carla.Vector3D) 
		"""
		...



class ApplyAngularImpulse:
	"""
	Command adaptation of add_angular_impulse() in carla.Actor. Applies an angular impulse to an actor.
	"""

	actor_id: int
	"""Actor affected by the command."""

	impulse: carla.Vector3D
	"""Angular impulse applied to the actor."""

	def __init__(self, actor: str, impulse: carla.Vector3D):
		"""


		:param actor: (str) Actor or its ID to whom the command will be applied to.

		:param impulse: (carla.Vector3D) 
		"""
		...



class ApplyTorque:
	"""
	Command adaptation of add_torque() in carla.Actor. Applies a torque to an actor.
	"""

	actor_id: int
	"""Actor affected by the command."""

	torque: carla.Vector3D
	"""Torque applied to the actor over time."""

	def __init__(self, actor: str, torque: carla.Vector3D):
		"""


		:param actor: (str) Actor or its ID to whom the command will be applied to.

		:param torque: (carla.Vector3D) 
		"""
		...



class SetSimulatePhysics:
	"""
	Command adaptation of set_simulate_physics() in carla.Actor. Determines whether an actor will be affected by physics or not.
	"""

	actor_id: int
	"""Actor affected by the command."""

	enabled: bool
	"""If physics should be activated or not."""

	def __init__(self, actor: str, enabled: bool):
		"""


		:param actor: (str) Actor or its ID to whom the command will be applied to.

		:param enabled: (bool) 
		"""
		...



class SetAutopilot:
	"""
	Command adaptation of set_autopilot() in carla.Vehicle. Turns on/off the vehicle's autopilot mode.
	"""

	actor_id: int
	"""Actor that is affected by the command."""

	enabled: bool
	"""If autopilot should be activated or not."""

	port: np.uint16
	"""Port of the Traffic Manager where the vehicle is to be registered or unlisted."""

	def __init__(self, actor: str, enabled: bool, port: np.uint16 = 8000):
		"""


		:param actor: (str) Actor or its ID to whom the command will be applied to.

		:param enabled: (bool) 

		:param port: (np.uint16) The Traffic Manager port where the vehicle is to be registered or unlisted. If __None__ is passed, it will consider a TM at default port `8000`.
		"""
		...



class SetVehicleLightState:
	"""
	Command adaptation of set_light_state() in carla.Vehicle. Sets the light state of a vehicle.
	"""

	actor_id: int
	"""Actor that is affected by the command."""

	light_state: carla.VehicleLightState
	"""Defines the light state of a vehicle."""

	def __init__(self, actor: str, light_state: carla.VehicleLightState):
		"""


		:param actor: (str) Actor or its ID to whom the command will be applied to.

		:param light_state: (carla.VehicleLightState) Recaps the state of the lights of a vehicle, these can be used as a flags.
		"""
		...



class SetEnableGravity:
	"""
	Command adaptation of set_enable_gravity() in carla.Actor. Enables or disables gravity on an actor.
	"""

	actor_id: str
	"""Actor that is affected by the command."""

	enabled: bool
	""""""

	def __init__(self, actor: str, enabled: bool):
		"""


		:param actor: (str) Actor or Actor ID to which the command will be applied to.

		:param enabled: (bool) 
		"""
		...



class ShowDebugTelemetry:
	"""
	Command adaptation of show_debug_telemetry() in carla.Actor. Displays vehicle control telemetry data.
	"""

	actor_id: str
	"""Actor that is affected by the command."""

	enabled: bool
	""""""

	def __init__(self, actor: str, enabled: bool):
		"""


		:param actor: (str) Actor or Actor ID to which the command will be applied to.

		:param enabled: (bool) 
		"""
		...






import warnings
warnings.filterwarnings(
    "ignore",
    message="pkg_resources is deprecated as an API",
    category=UserWarning
)

import io
import math
import random
import sys
from time import sleep
import carla
import click
import typer
from typing import Optional, Annotated
from . import utils, carla_utils, carla_game
import numpy as np
from pathlib import Path
import autonomous_proto
from rosbags.rosbag2 import Writer
from rosbags.typesys import Stores, get_typestore
from rosbags.typesys.stores.ros2_humble import (
    std_msgs__msg__UInt8MultiArray as UInt8MultiArray,
    std_msgs__msg__MultiArrayLayout as MultiArrayLayout,
    std_msgs__msg__MultiArrayDimension as MultiArrayDimension,
)

client: carla.Client = carla.Client('127.0.0.1', 2000)
app = typer.Typer(invoke_without_command=True)

@app.callback()
def main_callback(
        ctx: typer.Context,
        host: Annotated[str, typer.Option(help='CARLA host')] = '127.0.0.1',
        port: Annotated[int, typer.Option(help='CARLA port')] = 2000,
        timeout: Annotated[float, typer.Option(help='CARLA timeout')] = 1.0,
):
    global client
    client = carla.Client(host, port)
    client.set_timeout(timeout)
    if not ctx.invoked_subcommand:
        click.echo(ctx.get_help())

@app.command(name='print_actors', help='Print actors tree containing pattern')
def print_actors(pattern: Annotated[str, typer.Argument(help='pattern match the role_name or type_id')] = 'vehicle'):
    global client
    client.get_world().wait_for_tick()
    actors = client.get_world().get_actors()
    carla_utils.print_actors_tree(actors, pattern)

@app.command(name='print_actor', help='Print actor specified by actor id or role_name')
def print_actor(
        actor_id: Annotated[str, typer.Argument(help='Actor id or role_name')],
        loop: Annotated[bool, typer.Option(help='Loop mode')] = False,
):
    global client
    client.get_world().wait_for_tick()
    actor = carla_utils.get_actor(actor_id, client)
    if actor is None:
        print('Actor not found')
        return
    if not loop:
        carla_utils.print_actor(actor, client)
        return
    utils.set_stdout_nonblocking()
    while True:
        snapshot = client.get_world().wait_for_tick()
        origin_stdout = sys.stdout
        buffer = io.StringIO()
        sys.stdout = buffer
        sys.stdout.write("\x1b[H\x1b[2J\x1b[3J")
        sys.stdout.flush()
        print('CARLA Frame: ', snapshot.frame)
        carla_utils.print_actor(actor,
                                client,
                                print_parent = False,
                                print_sons = False,
                                print_attributes = False,
                                print_location = True,
                                print_rotation = True,
                                print_move = True,
                                print_boundbox = False,
                                print_vehicle_wheel = True,
                                print_vehicle_control = False,
                                print_vehicle_physics = False,
                                )
        sys.stdout = origin_stdout
        sys.stdout.write(buffer.getvalue())

@app.command(name='destroy_actor', help='Destroy actor with the id')
def destroy_actor(actor_id: Annotated[str, typer.Argument(help='The actor id or role_name of the actor to destroy.')]):
    global client
    client.get_world().wait_for_tick()
    carla_utils.destroy_actor(actor_id, client)

@app.command(name='available_maps', help='Print available maps')
def available_maps():
    global client
    client.get_world().wait_for_tick()
    print(*client.get_available_maps(), sep='\n')
    current_map = client.get_world().get_map()
    print()
    print('Current Map: ', current_map)

@app.command(name='change_map', help='Change map')
def change_map(map_name: Annotated[str, typer.Argument(help='The new map name.')] = None):
    global client
    client.set_timeout(10)
    client.get_world().wait_for_tick()
    current_map = client.get_world().get_map()
    print('Current Map:', current_map)
    if map_name is not None:
        client.load_world_if_different(map_name, False)
        print('Change map to:', map_name)
        return
    maps = client.get_available_maps()
    for i, m in enumerate(maps):
        print('{}:'.format(i), m)
    while True:
        index = input('index of map: ')
        try:
            index = int(index)
            if 0 <= index < len(maps):
                break
        except ValueError:
            pass
    client.load_world_if_different(maps[index], False)
    print('Change map to:', maps[index])


@app.command(name='available_bp', help='Print available blueprints')
def available_bp(pattern: Annotated[Optional[str], typer.Option(help="pattern match the blueprint type")] = '*'):
    global client
    client.get_world().wait_for_tick()
    if not pattern.startswith('*'):
        pattern = '*' + pattern
    if not pattern.endswith('*'):
        pattern = pattern + '*'
    print(*client.get_world().get_blueprint_library().filter(pattern), sep='\n')

@app.command(name='print_settings', help='Print current settings')
def print_settings():
    global client
    client.get_world().wait_for_tick()
    settings = client.get_world().get_settings()
    carla_utils.print_settings(settings)

@app.command(name='move', help='Move target to ...')
def move(
        actor_id: Annotated[str, typer.Argument(help='id of the actor to be move')] = 'spectator',
        to: Annotated[str, typer.Option(help='id of the actor move to')] = None,
        x: Annotated[float, typer.Option()] = None,
        y: Annotated[float, typer.Option()] = None,
        z: Annotated[float, typer.Option()] = None,
        roll: Annotated[float, typer.Option(help='degree')] = None,
        pitch: Annotated[float, typer.Option(help='degree')] = None,
        yaw: Annotated[float, typer.Option(help='degree')] = None,
):
    global client
    client.get_world().wait_for_tick()
    carla_utils.move_to(client, actor_id, to, x, y, z, roll, pitch, yaw)

@app.command(name='spy', help='Spy the specified actor')
def spy(actor_id: Annotated[str, typer.Argument(help='id of target actor')] = 'hero'):
    global client
    client.get_world().wait_for_tick()
    cg = carla_game.GUI(actor_id, client)
    cg.run()

@app.command(name='follow', help='Follow the spectator to actor')
def follow(
        actor_id: Annotated[str, typer.Argument(help='id of target actor')] = 'hero',
        x: Annotated[float, typer.Option()] = None,
        y: Annotated[float, typer.Option()] = None,
        z: Annotated[float, typer.Option()] = None,
        roll: Annotated[float, typer.Option(help='degree')] = None,
        pitch: Annotated[float, typer.Option(help='degree')] = None,
        yaw: Annotated[float, typer.Option(help='degree')] = None,
):
    global client
    client.get_world().wait_for_tick()
    actor = carla_utils.get_actor(actor_id, client)
    if actor is None or not actor.is_alive:
        print('Actor not found')
        return
    dummy_bp = carla_utils.get_blueprint('sensor.other.collision', client)
    dummy_bp.set_attribute('role_name', 'dummy_follow')
    dummy_transform = carla_utils.get_transform(x, y, z, roll, pitch, yaw)
    if dummy_transform is None and math.isfinite(actor.bounding_box.location.x):
        dummy_transform = carla.Transform(
            carla.Location(x=-actor.bounding_box.extent.x * 1.88, z=actor.bounding_box.extent.z * 3.76),
            carla.Rotation(pitch=-20)
        )
    dummy = carla_utils.spawn_actor(client, dummy_bp, dummy_transform, actor, carla.AttachmentType.Rigid)
    def stop():
        if dummy is not None and dummy.is_alive:
            dummy.destroy()
    world = client.get_world()
    world.wait_for_tick(1)
    spectator = world.get_spectator()
    try:
        while dummy.is_alive:
            world.wait_for_tick(1)
            spectator.set_transform(dummy.get_transform())
    except KeyboardInterrupt:
        stop()
    except RuntimeError:
        stop()
    finally:
        stop()

@app.command(name='pose_debug', help='Pose debug')
def pose_debug():
    global client
    client.get_world().wait_for_tick()
    spectator = carla_utils.get_actor('spectator', client)
    gnss_bp = carla_utils.get_blueprint('sensor.other.gnss', client)
    imu_bp = carla_utils.get_blueprint('sensor.other.imu', client)
    gnss: carla.Sensor = carla_utils.spawn_actor(client, gnss_bp, parent=spectator)
    imu: carla.Sensor = carla_utils.spawn_actor(client, imu_bp, parent=spectator)
    data = {'gnss': None, 'imu': None}
    gnss.listen(lambda e: data.update({'gnss': e}))
    imu.listen(lambda e: data.update({'imu': e}))
    try:
        while True:
            client.get_world().wait_for_tick()
            gnss_data: carla.GnssMeasurement = data['gnss']
            imu_data: carla.IMUMeasurement = data['imu']
            if gnss_data is None or imu_data is None:
                continue
            print('Gnss:')
            print('  Frame  :', gnss_data.frame)
            print('  lat    :', gnss_data.latitude)
            print('  lon    :', gnss_data.longitude)
            print('  alt    :', gnss_data.altitude)
            print('  X      :', gnss_data.transform.location.x)
            print('  Y      :', gnss_data.transform.location.y)
            print('  Z      :', gnss_data.transform.location.z)
            print('  roll   :', gnss_data.transform.rotation.roll)
            print('  pitch  :', gnss_data.transform.rotation.pitch)
            print('  yaw    :', gnss_data.transform.rotation.yaw)
            print('Imu:')
            print('  Frame  :', imu_data.frame)
            print('  compass:', math.degrees(imu_data.compass))
            print('  gyro.x :', imu_data.gyroscope.x)
            print('  gyro.y :', imu_data.gyroscope.y)
            print('  gyro.z :', imu_data.gyroscope.z)
            print('  acc.x  :', imu_data.accelerometer.x)
            print('  acc.y  :', imu_data.accelerometer.y)
            print('  acc.z  :', imu_data.accelerometer.z)
            print()
    finally:
        gnss.stop()
        imu.stop()
        carla_utils.destroy_actor(gnss.id, client)
        carla_utils.destroy_actor(imu.id, client)

@app.command(name='spawn_points', help='Print available spawn points')
def spawn_points():
    global client
    client.get_world().wait_for_tick()
    print(*client.get_world().get_map().get_spawn_points(), sep='\n')

@app.command(name='reset_to_spawn', help='Reset actor to spawn points')
def reset_to_spawn(actor_id: Annotated[str, typer.Argument(help='Actor id or role_name')] = 'hero'):
    global client
    client.get_world().wait_for_tick()
    actor = carla_utils.get_actor(actor_id, client)
    if actor is None:
        print('Actor {} not found'.format(actor_id))
        return
    points = client.get_world().get_map().get_spawn_points()
    if len(points) == 0:
        print('No spawn points')
        return
    for i, p in enumerate(points):
        print('{:>2d}. {}'.format(i, p))
    while True:
        index = input('index of spawn point: ')
        try:
            index = int(index)
            if 0 <= index < len(points):
                break
        except ValueError:
            pass
    actor.set_transform(points[index])
    client.get_world().get_spectator().set_transform(points[index])
    print('Reset actor to:', points[index])

@app.command(name='overlook', help='Overlook the whole map')
def overlook():
    global client
    client.get_world().wait_for_tick()
    waypoints = client.get_world().get_map().generate_waypoints(1)
    x_min = min(w.transform.location.x for w in waypoints)
    x_max = max(w.transform.location.x for w in waypoints)
    y_min = min(w.transform.location.y for w in waypoints)
    y_max = max(w.transform.location.y for w in waypoints)
    x_center = (x_max + x_min) / 2
    y_center = (y_max + y_min) / 2
    x_range = x_max - x_min
    y_range = y_max - y_min
    z = max(x_range, y_range)
    yaw = 0 if x_range < y_range else -90
    spectator = carla_utils.get_actor('spectator', client)
    spectator.set_transform(carla.Transform(carla.Location(x_center, y_center, z), carla.Rotation(pitch=-90, yaw=yaw)))
    print('x_min:', x_min)
    print('x_max:', x_max)
    print('x_center:', x_center)
    print('x_range:', x_range)
    print('y_min:', y_min)
    print('y_max:', y_max)
    print('y_center:', y_center)
    print('y_range:', y_range)
    print('z:', z)

@app.command(name='control_roadnet', help='Generate roadnet for control')
def control_roadnet(
        actor_id: Annotated[str, typer.Argument(help='Actor id or role_name')] = 'spectator',
        interval: Annotated[float, typer.Option] = 0.2,
        slow: Annotated[bool, typer.Option] = False,
):
    global client
    client.get_world().wait_for_tick()
    debug = client.get_world().debug
    spectator = client.get_world().get_spectator()
    world_map = client.get_world().get_map()
    actor = carla_utils.get_actor(actor_id, client)
    if actor is None:
        print('Actor not found')
        return
    start_point = world_map.get_waypoint(actor.get_location())
    if start_point is None:
        print('start point not found')
        return
    roadnet = []

    colors = [
        carla.Color(255, 0, 0),
        carla.Color(0, 255, 0),
        carla.Color(0, 0, 255),
    ]
    colors_str = ['red', 'green', 'blue']

    def draw_point(p: carla.Waypoint, z = 1.0, size: float = 0.1, c = carla.Color(0, 0, 0), lifetime = 0):
        t = p.transform.location
        t.z += z
        debug.draw_point(t, size=size, color=c, life_time=lifetime)
    def get_line(s: carla.Waypoint):
        c = s
        points = [c]
        while True:
            w = c.next(interval)
            if len(w) != 1:
                return points
            c = w[0]
            points.append(c)
    def move(p: carla.Waypoint):
        draw_point(p, size=0.3)
        t = p.transform
        t.location.z += 2
        t.rotation.pitch = -20
        t.location.x -= 2 * math.cos(math.radians(t.rotation.yaw))
        t.location.y -= 2 * math.sin(math.radians(t.rotation.yaw))
        spectator.set_transform(t)
        if slow:
            input('press enter to add current point')
        else:
            sleep(0.02)

    try:
        while True:
            next_points = start_point.next(interval)
            if len(next_points) == 0 or len(next_points) > 2:
                print('next_points(roads) size:', len(next_points))
                break
            if len(next_points) == 1:
                move(next_points[0])
                start_point = next_points[0]
                roadnet.append(start_point)
                continue
            for i, point in enumerate(next_points):
                print('{}:'.format(i), colors_str[i])
                for p in get_line(point):
                    draw_point(p, z=1.2, c=colors[i], lifetime=10)

            index = input('index of selected: ')
            try:
                index = int(index)
                if 0 <= index < len(next_points):
                    move(next_points[index])
                    start_point = next_points[index]
                    roadnet.append(start_point)
            except ValueError:
                pass
            move(start_point)
    except KeyboardInterrupt:
        print('keyboard interrupt')

    print('collected roadnet points size:', len(roadnet))
    def get_interval(a: carla.Waypoint, b: carla.Waypoint):
        x = a.transform.location.x - b.transform.location.x
        y = a.transform.location.y - b.transform.location.y
        return math.sqrt(x * x + y * y)
    min_interval = min(get_interval(a, b) for a, b in zip(roadnet, roadnet[1:]))
    max_interval = max(get_interval(a, b) for a, b in zip(roadnet, roadnet[1:]))
    print('min_interval:', min_interval)
    print('max_interval:', max_interval)

    coordinates = [world_map.transform_to_geolocation(t.transform.location) for t in roadnet]
    navs = []
    nav = autonomous_proto.Navigation()
    nav.header.info.count = 0
    nav.header.info.module_name = autonomous_proto.MessageInfoModuleNameValue.undefined
    nav.header.info.topic_name = autonomous_proto.MessageInfoTopicNameValue.undefined
    for coordinate in coordinates:
        nav.header.info.count += 1
        nav.position.lat = coordinate.latitude
        nav.position.lon = coordinate.longitude
        nav.position.alt = coordinate.altitude
        navs.append(nav.SerializeToString())

    bag_path = Path(input('bag name: '))
    with Writer(bag_path, version=8) as writer:
        type_store = get_typestore(Stores.ROS2_HUMBLE)
        topic = 'navigation'
        msg_type = UInt8MultiArray.__msgtype__
        connection = writer.add_connection(topic, msg_type, typestore=type_store)
        timestamp = 0
        for n in navs:
            timestamp += 1
            data = np.frombuffer(n, dtype=np.uint8)
            dims = [MultiArrayDimension(label='', size=data.size, stride=1)]
            layout = MultiArrayLayout(dim=dims, data_offset=0)
            message = UInt8MultiArray(layout=layout, data=data)
            writer.write(connection, timestamp, type_store.serialize_cdr(message, msg_type))



# @app.command(name='test')
# def test():
#     global client
#     client.get_world().wait_for_tick()

if __name__ in "__main__":
    try:
        app()
    except RuntimeError as e:
        print(e)
